==> allensdk-1.3.2-py3-none-any.whl: allensdk/brain_observatory/behavior/swdb/save_flash_response_df.py <==
import sys
import os
import numpy as np
import pandas as pd
import itertools

from allensdk.brain_observatory.behavior.behavior_ophys_session import BehaviorOphysSession
from allensdk.brain_observatory.behavior.behavior_ophys_api.behavior_ophys_nwb_api import BehaviorOphysNwbApi
from allensdk.internal.api.behavior_ophys_api import BehaviorOphysLimsApi
from allensdk.brain_observatory.behavior.swdb import behavior_project_cache as bpc
from allensdk.brain_observatory.behavior.swdb.analysis_tools import get_nearest_frame, get_trace_around_timepoint, get_mean_in_window

'''
    This script computes the flash_response_df for a BehaviorOphysSession object

'''

def get_flash_response_df(session, response_analysis_params):
    '''
        Builds the flash response dataframe for <session>

        INPUTS:
        <session> BehaviorOphysSession to build the flash response dataframe for
        <response_analyis_params>   A dictionary with the following keys 
                'window_around_timepoint_seconds' is the time window to save out the dff_trace around the flash onset.
                'response_window_duration_seconds' is the length of time after the flash onset to compute the mean_response
                'baseline_window_duration_seconds' is the length of time before the flash onset to compute the baseline response
        
        OUTPUTS:
        A dataframe with index: (cell_specimen_id, flash_id)
        and columns:
        cell_roi_id, the cell's roi id for that session
        mean_response, the mean df/f in the response_window
        baseline_response, the mean df/f in the baseline_window
        dff_trace, the dff trace in the window_around_timepoint_seconds
        dff_trace_timestamps, the timestamps for the dff_trace

    '''
    frame_rate = 31.    # Shouldn't hard code this here

    # get data to analyze
    dff_traces = session.dff_traces.copy()
    flashes = session.stimulus_presentations.copy()

    # get params to define response window, in seconds
    window_around_timepoint_seconds = response_analysis_params['window_around_timepoint_seconds']
    response_window_duration_seconds = response_analysis_params['response_window_duration_seconds']
    baseline_window_duration_seconds = response_analysis_params['baseline_window_duration_seconds']
    mean_response_window_seconds = [np.abs(window_around_timepoint_seconds[0]), 
                        np.abs(window_around_timepoint_seconds[0]) + response_window_duration_seconds]
    baseline_window_seconds = [np.abs(window_around_timepoint_seconds[0]) - baseline_window_duration_seconds, 
                        np.abs(window_around_timepoint_seconds[0])]

    # Build a dataframe with multiindex defined as product of cell_id X flash_id
    cell_flash_combinations = itertools.product(dff_traces.index,flashes.index)
    index = pd.MultiIndex.from_tuples(cell_flash_combinations, names=['cell_specimen_id', 'flash_id'])
    df = pd.DataFrame(index=index)
    traces_list = []
    trace_timestamps_list = []
    
    # Iterate though cell/flash pairs and build table
    for cell_specimen_id, flash_id in itertools.product(dff_traces.index,flashes.index):
        timepoint = flashes.loc[flash_id]['start_time']
        cell_roi_id = dff_traces.loc[cell_specimen_id]['cell_roi_id']
        full_cell_trace = dff_traces.loc[cell_specimen_id, 'dff']
        trace, trace_timestamps = get_trace_around_timepoint(full_cell_trace, timepoint, session.ophys_timestamps, window_around_timepoint_seconds, frame_rate)
        mean_response = get_mean_in_window(trace, mean_response_window_seconds, frame_rate)
        baseline_response = get_mean_in_window(trace, baseline_window_seconds, frame_rate) 
        traces_list.append(trace)
        trace_timestamps_list.append(trace_timestamps)
        df.loc[(cell_specimen_id, flash_id), 'cell_roi_id'] = int(cell_roi_id)
        df.loc[(cell_specimen_id, flash_id), 'mean_response'] = mean_response
        df.loc[(cell_specimen_id, flash_id), 'baseline_response'] = baseline_response
    df.insert(loc=1, column='dff_trace', value=traces_list)
    df.insert(loc=2, column='dff_trace_timestamps', value=trace_timestamps_list)
    return df

def get_p_values_from_shuffled_spontaneous(session, flash_response_df, response_window_duration=0.5,number_of_shuffles=10000):
    '''
        Computes the P values for each cell/flash. The P value is the probability of observing a response of that
        magnitude in the spontaneous window. The algorithm is copied from VBA

        INPUTS:
            <session> a BehaviorOphysSession object
            <flash_response_df> the flash_response_df for this session
            <response_window_duration> is the duration of the response_window that was used to compute the mean_response in the flash_response_df. This is used here to extract an equivalent duration df/f trace from the spontaneous timepoint
            <number_of_shuffles> the number of shuffles of spontaneous activity used to compute the pvalue
    
        OUTPUTS:
        fdf, a copy of the flash_response_df with a new column appended 'p_value' which is the per-flash X per-cell p-value
    
        ASSERTS:
            each p value is bounded by 0 and 1, and does not include any NaNs

    '''
    # Organize Data
    fdf = flash_response_df.copy()
    st  = session.stimulus_presentations.copy()
    included_flashes = fdf.index.get_level_values(1).unique()
    st  = st[st.index.isin(included_flashes)]

    # Get Sample of Spontaneous Frames
    spontaneous_frames = get_spontaneous_frames(session)

    #  Compute the number of response_window frames
    ophys_frame_rate = 31 # Shouldn't hard code this here
    n_mean_response_window_frames = int(np.round(response_window_duration * ophys_frame_rate, 0))
    cell_ids = np.unique(fdf.index.get_level_values(0))    
    n_cells = len(cell_ids)

    # Get Shuffled responses from spontaneous frames
    # get mean response for shuffles of the spontaneous activity frames
    # in a window the same size as the stim response window duration
    shuffled_responses = np.empty((n_cells, number_of_shuffles, n_mean_response_window_frames))
    idx = np.random.choice(spontaneous_frames, number_of_shuffles)
    dff_traces = np.stack(session.dff_traces.to_numpy()[:,1],axis=0)
    for i in range(n_mean_response_window_frames):
        shuffled_responses[:, :, i] = dff_traces[:, idx + i]
    shuffled_mean = shuffled_responses.mean(axis=2)

    # compare flash responses to shuffled values and make a dataframe of p_value for cell_id X flash_id
    iterables = [cell_ids, st.index.values]
    flash_p_values = pd.DataFrame(index=pd.MultiIndex.from_product(iterables,names= ['cell_specimen_id','flash_id']))
    for i, cell_index in enumerate(cell_ids):
        responses = fdf.loc[cell_index].mean_response.values
        null_dist_mat = np.tile(shuffled_mean[i, :], reps=(len(responses), 1))
        actual_is_less = responses.reshape(len(responses), 1) <= null_dist_mat
        p_values = np.mean(actual_is_less, axis=1)
        for j in range(0,len(p_values)):
            flash_p_values.at[(cell_index,j),'p_value'] = p_values[j] 
    fdf = pd.concat([fdf,flash_p_values],axis=1)

    # Test to ensure p values are bounded between 0 and 1, and dont include NaNs
    assert np.all(fdf['p_value'].values <= 1)
    assert np.all(fdf['p_value'].values >= 0)
    assert np.all(~np.isnan(fdf['p_value'].values))
    
    return fdf 

def get_spontaneous_frames(session):
    '''
        Returns a list of the frames that occur during the before and after spontaneous windows. This is copied from VBA. Does not use the full spontaneous period because that is what VBA did. It only uses 4 minutes of the before and after spontaneous period. 
    
        INPUTS:
        <session> a BehaviorOphysSession object to get all the spontaneous frames 

        OUTPUTS: a list of the frames during the spontaneous period
    '''
    st = session.stimulus_presentations.copy()
    # dont use full 5 mins to avoid fingerprint and countdown
    # spont_duration_frames = 4 * 60 * 60  # 4 mins * * 60s/min * 60Hz
    spont_duration = 4 * 60  # 4mins * 60sec

    # for spontaneous at beginning of session
    behavior_start_time = st.iloc[0].start_time
    spontaneous_start_time_pre = behavior_start_time - spont_duration
    spontaneous_end_time_pre = behavior_start_time
    spontaneous_start_frame_pre = get_successive_frame_list(spontaneous_start_time_pre, session.ophys_timestamps)
    spontaneous_end_frame_pre = get_successive_frame_list(spontaneous_end_time_pre, session.ophys_timestamps)
    spontaneous_frames_pre = np.arange(spontaneous_start_frame_pre, spontaneous_end_frame_pre, 1)

    # for spontaneous epoch at end of session
    behavior_end_time = st.iloc[-1].stop_time
    spontaneous_start_time_post = behavior_end_time + 0.5
    spontaneous_end_time_post = behavior_end_time + spont_duration
    spontaneous_start_frame_post = get_successive_frame_list(spontaneous_start_time_post, session.ophys_timestamps)
    spontaneous_end_frame_post = get_successive_frame_list(spontaneous_end_time_post, session.ophys_timestamps)
    spontaneous_frames_post = np.arange(spontaneous_start_frame_post, spontaneous_end_frame_post, 1)

    # add them together
    spontaneous_frames = list(spontaneous_frames_pre) + (list(spontaneous_frames_post))
    return spontaneous_frames

def get_successive_frame_list(timepoints_array, timestamps):
    '''
        Returns the next frame after timestamps in timepoints_array
        copied from VBA
    '''
    # This is a modification of get_nearest_frame for speedup
    #  This implementation looks for the first 2p frame consecutive to the stim
    successive_frames = np.searchsorted(timestamps, timepoints_array)
    return successive_frames

def add_image_name(session,fdf):
    '''
        Adds a column to flash_response_df with the image_name taken from the stimulus_presentations table
        Slow to run, could probably be improved with some more intelligent use of pandas
    
        INPUTS:
        <session> a BehaviorOphysSession object
        <fdf> a flash_response_df for this session

        OUTPUTS:
        fdf, with a new column appended 'image_name' which gives the image identity (like 'im066') for each flash.
    '''

    fdf = fdf.reset_index()
    fdf = fdf.set_index('flash_id')
    fdf['image_name']= ''
    # So slow!!!
    for stim_id in np.unique(fdf.index.values):
        fdf.loc[stim_id,'image_name'] = session.stimulus_presentations.loc[stim_id].image_name
    fdf = fdf.reset_index()
    fdf = fdf.set_index(['cell_specimen_id','flash_id'])
    return fdf

def annotate_flash_response_df_with_pref_stim(fdf):
    '''
        Adds a column to flash_response_df with a boolean value of whether that flash was that cells pref image.
        Computes preferred image by looking for the image that on average evokes the largest response.
        Slow to run, could probably be improved with more intelligent pandas use

        INPUTS:
        fdf, a flash_response_dataframe

        RETURNS:
        fdf, appended with 'pref_stim' column

        ASSERTS:
        each cell has one unique preferred_stimulus

    '''
    # Prepare dataframe
    fdf = fdf.reset_index()
    if 'cell_specimen_id' in fdf.keys():
        cell_key = 'cell_specimen_id'
    else:
        cell_key = 'cell'
    
    # Set up empty column
    fdf['pref_stim'] = False
    
    # Compute average response for each image
    mean_response = fdf.groupby([cell_key, 'image_name']).apply(get_mean_sem)
    m = mean_response.unstack()
    
    # Iterate through each cell and find which image evoked the largest average response
    for cell in m.index:
        temp = np.where(m.loc[cell]['mean_response'].values == np.nanmax(m.loc[cell]['mean_response'].values))[0]
        # If the mean_response was NaN, then temp is empty, so we have this check here
        if len(temp) > 0:
            image_index = temp[0]
            pref_image = m.loc[cell]['mean_response'].index[image_index]
            # find all repeats of that cell X pref_image, and set 'pref_stim' to True
            cell_flash_pairs = fdf[(fdf[cell_key] == cell) & (fdf.image_name == pref_image)].index
            fdf.loc[cell_flash_pairs, 'pref_stim'] = True

    # Test to ensure preferred stimulus is unique for each cell
    for cell in fdf['cell_specimen_id'].unique():
        assert len(fdf.set_index('cell_specimen_id').loc[cell].query('pref_stim').image_name.unique()) == 1

    # Reset the df index
    fdf = fdf.set_index(['cell_specimen_id','flash_id'])
    return fdf

def get_mean_sem(group):
    '''
        Returns the mean and sem of the mean_response values for all entries in the group. Copied from VBA
        
        INPUTS:
        group is a pandas group
        
        Output, a pandas series with the average 'mean_response' from the group, and the sem 'mean_response' from the group
    '''
    mean_response = np.mean(group['mean_response'])
    sem_response = np.std(group['mean_response'].values) / np.sqrt(len(group['mean_response'].values))
    return pd.Series({'mean_response': mean_response, 'sem_response': sem_response})


if __name__=='__main__':

    case = 0

    if case==0:
        # This is the main usage case. 
    
        # Grab the experiment ID
        experiment_id = sys.argv[1]

        # Define the cache
        cache_json = {'manifest_path': '/allen/programs/braintv/workgroups/nc-ophys/visual_behavior/SWDB_2019/visual_behavior_data_manifest.csv',
                      'nwb_base_dir': '/allen/programs/braintv/workgroups/nc-ophys/visual_behavior/SWDB_2019/nwb_files',
                      'analysis_files_base_dir': '/allen/programs/braintv/workgroups/nc-ophys/visual_behavior/SWDB_2019/extra_files'
                      }

        # load the session
        cache = bpc.BehaviorProjectCache(cache_json)
        nwb_path = cache.get_nwb_filepath(experiment_id)
        api = BehaviorOphysNwbApi(nwb_path, filter_invalid_rois = True)
        session = BehaviorOphysSession(api)

        # Where to save the results
        output_path = '/allen/programs/braintv/workgroups/nc-ophys/visual_behavior/SWDB_2019/flash_response_500msec_response'

        # Define parameters for dff_trace, and response_window
        response_analysis_params = {'window_around_timepoint_seconds':[-.5,.75], # -500ms, 750ms
                                   'response_window_duration_seconds': 0.5, 
                                   'baseline_window_duration_seconds': 0.5} 

        # compute the base flash_response_df
        flash_response_df = get_flash_response_df(session, response_analysis_params)

        # Add p_value, image_name, and pref_stim
        flash_response_df = get_p_values_from_shuffled_spontaneous(session,flash_response_df)
        flash_response_df = add_image_name(session,flash_response_df)
        flash_response_df = annotate_flash_response_df_with_pref_stim(flash_response_df)
    
        # Test columns in flash_response_df
        for new_key in ['cell_roi_id','mean_response','baseline_response','dff_trace','dff_trace_timestamps','p_value','image_name', 'pref_stim']:
            assert new_key in flash_response_df.keys()

        # Save the flash_response_df to file
        output_fn = os.path.join(output_path, 'flash_response_df_{}.h5'.format(experiment_id))
        print('Writing flash response df to {}'.format(output_fn))
        flash_response_df.to_hdf(output_fn, key='df', complib='bzip2', complevel=9)

    elif case==1:
        # This case is just for debugging. It computes the flash_response_df on a truncated portion of the data. 
        nwb_path = '/allen/programs/braintv/workgroups/nc-ophys/visual_behavior/SWDB_2019/nwb_files/behavior_ophys_session_880961028.nwb'
        api = BehaviorOphysNwbApi(nwb_path, filter_invalid_rois=True)
        session = BehaviorOphysSession(api)

        #Small data for testing
        session.__dict__['dff_traces'].value = session.dff_traces.iloc[:5]
        session.__dict__['stimulus_presentations'].value = session.stimulus_presentations.iloc[:20]

        response_analysis_params = {'window_around_timepoint_seconds':[-.5,.75], # -500ms, 750ms
                                   'response_window_duration_seconds': 0.5, 
                                   'baseline_window_duration_seconds': 0.5} 

        flash_response_df = get_flash_response_df(session, response_analysis_params)
        flash_response_df = get_p_values_from_shuffled_spontaneous(session,flash_response_df)
        flash_response_df = add_image_name(session,flash_response_df)
        flash_response_df = annotate_flash_response_df_with_pref_stim(flash_response_df)

        # Test columns in flash_response_df
        for new_key in ['cell_roi_id','mean_response','baseline_response','dff_trace','dff_trace_timestamps','p_value','image_name', 'pref_stim']:
            assert new_key in flash_response_df.keys()

==> allensdk-2.8.0-py3-none-any.whl: allensdk/brain_observatory/behavior/swdb/save_flash_response_df.py <==
import sys
import os
import numpy as np
import pandas as pd
import itertools

from allensdk.brain_observatory.behavior.behavior_ophys_session import BehaviorOphysSession
from allensdk.brain_observatory.behavior.session_apis.data_io import (
    BehaviorOphysNwbApi)
from allensdk.brain_observatory.behavior.session_apis.data_io import (
    BehaviorOphysLimsApi)
from allensdk.brain_observatory.behavior.swdb import behavior_project_cache as bpc
from allensdk.brain_observatory.behavior.swdb.analysis_tools import get_nearest_frame, get_trace_around_timepoint, get_mean_in_window

'''
    This script computes the flash_response_df for a BehaviorOphysSession object

'''

def get_flash_response_df(session, response_analysis_params):
    '''
        Builds the flash response dataframe for <session>

        INPUTS:
        <session> BehaviorOphysSession to build the flash response dataframe for
        <response_analyis_params>   A dictionary with the following keys 
                'window_around_timepoint_seconds' is the time window to save out the dff_trace around the flash onset.
                'response_window_duration_seconds' is the length of time after the flash onset to compute the mean_response
                'baseline_window_duration_seconds' is the length of time before the flash onset to compute the baseline response
        
        OUTPUTS:
        A dataframe with index: (cell_specimen_id, flash_id)
        and columns:
        cell_roi_id, the cell's roi id for that session
        mean_response, the mean df/f in the response_window
        baseline_response, the mean df/f in the baseline_window
        dff_trace, the dff trace in the window_around_timepoint_seconds
        dff_trace_timestamps, the timestamps for the dff_trace

    '''
    frame_rate = 31.    # Shouldn't hard code this here

    # get data to analyze
    dff_traces = session.dff_traces.copy()
    flashes = session.stimulus_presentations.copy()

    # get params to define response window, in seconds
    window_around_timepoint_seconds = response_analysis_params['window_around_timepoint_seconds']
    response_window_duration_seconds = response_analysis_params['response_window_duration_seconds']
    baseline_window_duration_seconds = response_analysis_params['baseline_window_duration_seconds']
    mean_response_window_seconds = [np.abs(window_around_timepoint_seconds[0]), 
                        np.abs(window_around_timepoint_seconds[0]) + response_window_duration_seconds]
    baseline_window_seconds = [np.abs(window_around_timepoint_seconds[0]) - baseline_window_duration_seconds, 
                        np.abs(window_around_timepoint_seconds[0])]

    # Build a dataframe with multiindex defined as product of cell_id X flash_id
    cell_flash_combinations = itertools.product(dff_traces.index,flashes.index)
    index = pd.MultiIndex.from_tuples(cell_flash_combinations, names=['cell_specimen_id', 'flash_id'])
    df = pd.DataFrame(index=index)
    traces_list = []
    trace_timestamps_list = []
    
    # Iterate though cell/flash pairs and build table
    for cell_specimen_id, flash_id in itertools.product(dff_traces.index,flashes.index):
        timepoint = flashes.loc[flash_id]['start_time']
        cell_roi_id = dff_traces.loc[cell_specimen_id]['cell_roi_id']
        full_cell_trace = dff_traces.loc[cell_specimen_id, 'dff']
        trace, trace_timestamps = get_trace_around_timepoint(full_cell_trace, timepoint, session.ophys_timestamps, window_around_timepoint_seconds, frame_rate)
        mean_response = get_mean_in_window(trace, mean_response_window_seconds, frame_rate)
        baseline_response = get_mean_in_window(trace, baseline_window_seconds, frame_rate) 
        traces_list.append(trace)
        trace_timestamps_list.append(trace_timestamps)
        df.loc[(cell_specimen_id, flash_id), 'cell_roi_id'] = int(cell_roi_id)
        df.loc[(cell_specimen_id, flash_id), 'mean_response'] = mean_response
        df.loc[(cell_specimen_id, flash_id), 'baseline_response'] = baseline_response
    df.insert(loc=1, column='dff_trace', value=traces_list)
    df.insert(loc=2, column='dff_trace_timestamps', value=trace_timestamps_list)
    return df

def get_p_values_from_shuffled_spontaneous(session, flash_response_df, response_window_duration=0.5,number_of_shuffles=10000):
    '''
        Computes the P values for each cell/flash. The P value is the probability of observing a response of that
        magnitude in the spontaneous window. The algorithm is copied from VBA

        INPUTS:
            <session> a BehaviorOphysSession object
            <flash_response_df> the flash_response_df for this session
            <response_window_duration> is the duration of the response_window that was used to compute the mean_response in the flash_response_df. This is used here to extract an equivalent duration df/f trace from the spontaneous timepoint
            <number_of_shuffles> the number of shuffles of spontaneous activity used to compute the pvalue
    
        OUTPUTS:
        fdf, a copy of the flash_response_df with a new column appended 'p_value' which is the per-flash X per-cell p-value
    
        ASSERTS:
            each p value is bounded by 0 and 1, and does not include any NaNs

    '''
    # Organize Data
    fdf = flash_response_df.copy()
    st  = session.stimulus_presentations.copy()
    included_flashes = fdf.index.get_level_values(1).unique()
    st  = st[st.index.isin(included_flashes)]

    # Get Sample of Spontaneous Frames
    spontaneous_frames = get_spontaneous_frames(session)

    #  Compute the number of response_window frames
    ophys_frame_rate = 31 # Shouldn't hard code this here
    n_mean_response_window_frames = int(np.round(response_window_duration * ophys_frame_rate, 0))
    cell_ids = np.unique(fdf.index.get_level_values(0))    
    n_cells = len(cell_ids)

    # Get Shuffled responses from spontaneous frames
    # get mean response for shuffles of the spontaneous activity frames
    # in a window the same size as the stim response window duration
    shuffled_responses = np.empty((n_cells, number_of_shuffles, n_mean_response_window_frames))
    idx = np.random.choice(spontaneous_frames, number_of_shuffles)
    dff_traces = np.stack(session.dff_traces.to_numpy()[:,1],axis=0)
    for i in range(n_mean_response_window_frames):
        shuffled_responses[:, :, i] = dff_traces[:, idx + i]
    shuffled_mean = shuffled_responses.mean(axis=2)

    # compare flash responses to shuffled values and make a dataframe of p_value for cell_id X flash_id
    iterables = [cell_ids, st.index.values]
    flash_p_values = pd.DataFrame(index=pd.MultiIndex.from_product(iterables,names= ['cell_specimen_id','flash_id']))
    for i, cell_index in enumerate(cell_ids):
        responses = fdf.loc[cell_index].mean_response.values
        null_dist_mat = np.tile(shuffled_mean[i, :], reps=(len(responses), 1))
        actual_is_less = responses.reshape(len(responses), 1) <= null_dist_mat
        p_values = np.mean(actual_is_less, axis=1)
        for j in range(0,len(p_values)):
            flash_p_values.at[(cell_index,j),'p_value'] = p_values[j] 
    fdf = pd.concat([fdf,flash_p_values],axis=1)

    # Test to ensure p values are bounded between 0 and 1, and dont include NaNs
    assert np.all(fdf['p_value'].values <= 1)
    assert np.all(fdf['p_value'].values >= 0)
    assert np.all(~np.isnan(fdf['p_value'].values))
    
    return fdf 

def get_spontaneous_frames(session):
    '''
        Returns a list of the frames that occur during the before and after spontaneous windows. This is copied from VBA. Does not use the full spontaneous period because that is what VBA did. It only uses 4 minutes of the before and after spontaneous period. 
    
        INPUTS:
        <session> a BehaviorOphysSession object to get all the spontaneous frames 

        OUTPUTS: a list of the frames during the spontaneous period
    '''
    st = session.stimulus_presentations.copy()
    # dont use full 5 mins to avoid fingerprint and countdown
    # spont_duration_frames = 4 * 60 * 60  # 4 mins * * 60s/min * 60Hz
    spont_duration = 4 * 60  # 4mins * 60sec

    # for spontaneous at beginning of session
    behavior_start_time = st.iloc[0].start_time
    spontaneous_start_time_pre = behavior_start_time - spont_duration
    spontaneous_end_time_pre = behavior_start_time
    spontaneous_start_frame_pre = get_successive_frame_list(spontaneous_start_time_pre, session.ophys_timestamps)
    spontaneous_end_frame_pre = get_successive_frame_list(spontaneous_end_time_pre, session.ophys_timestamps)
    spontaneous_frames_pre = np.arange(spontaneous_start_frame_pre, spontaneous_end_frame_pre, 1)

    # for spontaneous epoch at end of session
    behavior_end_time = st.iloc[-1].stop_time
    spontaneous_start_time_post = behavior_end_time + 0.5
    spontaneous_end_time_post = behavior_end_time + spont_duration
    spontaneous_start_frame_post = get_successive_frame_list(spontaneous_start_time_post, session.ophys_timestamps)
    spontaneous_end_frame_post = get_successive_frame_list(spontaneous_end_time_post, session.ophys_timestamps)
    spontaneous_frames_post = np.arange(spontaneous_start_frame_post, spontaneous_end_frame_post, 1)

    # add them together
    spontaneous_frames = list(spontaneous_frames_pre) + (list(spontaneous_frames_post))
    return spontaneous_frames

def get_successive_frame_list(timepoints_array, timestamps):
    '''
        Returns the next frame after timestamps in timepoints_array
        copied from VBA
    '''
    # This is a modification of get_nearest_frame for speedup
    #  This implementation looks for the first 2p frame consecutive to the stim
    successive_frames = np.searchsorted(timestamps, timepoints_array)
    return successive_frames

def add_image_name(session,fdf):
    '''
        Adds a column to flash_response_df with the image_name taken from the stimulus_presentations table
        Slow to run, could probably be improved with some more intelligent use of pandas
    
        INPUTS:
        <session> a BehaviorOphysSession object
        <fdf> a flash_response_df for this session

        OUTPUTS:
        fdf, with a new column appended 'image_name' which gives the image identity (like 'im066') for each flash.
    '''

    fdf = fdf.reset_index()
    fdf = fdf.set_index('flash_id')
    fdf['image_name']= ''
    # So slow!!!
    for stim_id in np.unique(fdf.index.values):
        fdf.loc[stim_id,'image_name'] = session.stimulus_presentations.loc[stim_id].image_name
    fdf = fdf.reset_index()
    fdf = fdf.set_index(['cell_specimen_id','flash_id'])
    return fdf

def annotate_flash_response_df_with_pref_stim(fdf):
    '''
        Adds a column to flash_response_df with a boolean value of whether that flash was that cells pref image.
        Computes preferred image by looking for the image that on average evokes the largest response.
        Slow to run, could probably be improved with more intelligent pandas use

        INPUTS:
        fdf, a flash_response_dataframe

        RETURNS:
        fdf, appended with 'pref_stim' column

        ASSERTS:
        each cell has one unique preferred_stimulus

    '''
    # Prepare dataframe
    fdf = fdf.reset_index()
    if 'cell_specimen_id' in fdf.keys():
        cell_key = 'cell_specimen_id'
    else:
        cell_key = 'cell'
    
    # Set up empty column
    fdf['pref_stim'] = False
    
    # Compute average response for each image
    mean_response = fdf.groupby([cell_key, 'image_name']).apply(get_mean_sem)
    m = mean_response.unstack()
    
    # Iterate through each cell and find which image evoked the largest average response
    for cell in m.index:
        temp = np.where(m.loc[cell]['mean_response'].values == np.nanmax(m.loc[cell]['mean_response'].values))[0]
        # If the mean_response was NaN, then temp is empty, so we have this check here
        if len(temp) > 0:
            image_index = temp[0]
            pref_image = m.loc[cell]['mean_response'].index[image_index]
            # find all repeats of that cell X pref_image, and set 'pref_stim' to True
            cell_flash_pairs = fdf[(fdf[cell_key] == cell) & (fdf.image_name == pref_image)].index
            fdf.loc[cell_flash_pairs, 'pref_stim'] = True

    # Test to ensure preferred stimulus is unique for each cell
    for cell in fdf['cell_specimen_id'].unique():
        assert len(fdf.set_index('cell_specimen_id').loc[cell].query('pref_stim').image_name.unique()) == 1

    # Reset the df index
    fdf = fdf.set_index(['cell_specimen_id','flash_id'])
    return fdf

def get_mean_sem(group):
    '''
        Returns the mean and sem of the mean_response values for all entries in the group. Copied from VBA
        
        INPUTS:
        group is a pandas group
        
        Output, a pandas series with the average 'mean_response' from the group, and the sem 'mean_response' from the group
    '''
    mean_response = np.mean(group['mean_response'])
    sem_response = np.std(group['mean_response'].values) / np.sqrt(len(group['mean_response'].values))
    return pd.Series({'mean_response': mean_response, 'sem_response': sem_response})


if __name__=='__main__':

    case = 0

    if case==0:
        # This is the main usage case. 
    
        # Grab the experiment ID
        experiment_id = sys.argv[1]

        # Define the cache
        cache_json = {'manifest_path': '/allen/programs/braintv/workgroups/nc-ophys/visual_behavior/SWDB_2019/visual_behavior_data_manifest.csv',
                      'nwb_base_dir': '/allen/programs/braintv/workgroups/nc-ophys/visual_behavior/SWDB_2019/nwb_files',
                      'analysis_files_base_dir': '/allen/programs/braintv/workgroups/nc-ophys/visual_behavior/SWDB_2019/extra_files'
                      }

        # load the session
        cache = bpc.BehaviorProjectCache(cache_json)
        nwb_path = cache.get_nwb_filepath(experiment_id)
        api = BehaviorOphysNwbApi(nwb_path, filter_invalid_rois = True)
        session = BehaviorOphysSession(api)

        # Where to save the results
        output_path = '/allen/programs/braintv/workgroups/nc-ophys/visual_behavior/SWDB_2019/flash_response_500msec_response'

        # Define parameters for dff_trace, and response_window
        response_analysis_params = {'window_around_timepoint_seconds':[-.5,.75], # -500ms, 750ms
                                   'response_window_duration_seconds': 0.5, 
                                   'baseline_window_duration_seconds': 0.5} 

        # compute the base flash_response_df
        flash_response_df = get_flash_response_df(session, response_analysis_params)

        # Add p_value, image_name, and pref_stim
        flash_response_df = get_p_values_from_shuffled_spontaneous(session,flash_response_df)
        flash_response_df = add_image_name(session,flash_response_df)
        flash_response_df = annotate_flash_response_df_with_pref_stim(flash_response_df)
    
        # Test columns in flash_response_df
        for new_key in ['cell_roi_id','mean_response','baseline_response','dff_trace','dff_trace_timestamps','p_value','image_name', 'pref_stim']:
            assert new_key in flash_response_df.keys()

        # Save the flash_response_df to file
        output_fn = os.path.join(output_path, 'flash_response_df_{}.h5'.format(experiment_id))
        print('Writing flash response df to {}'.format(output_fn))
        flash_response_df.to_hdf(output_fn, key='df', complib='bzip2', complevel=9)

    elif case==1:
        # This case is just for debugging. It computes the flash_response_df on a truncated portion of the data. 
        nwb_path = '/allen/programs/braintv/workgroups/nc-ophys/visual_behavior/SWDB_2019/nwb_files/behavior_ophys_session_880961028.nwb'
        api = BehaviorOphysNwbApi(nwb_path, filter_invalid_rois=True)
        session = BehaviorOphysSession(api)

        #Small data for testing
        session.__dict__['dff_traces'].value = session.dff_traces.iloc[:5]
        session.__dict__['stimulus_presentations'].value = session.stimulus_presentations.iloc[:20]

        response_analysis_params = {'window_around_timepoint_seconds':[-.5,.75], # -500ms, 750ms
                                   'response_window_duration_seconds': 0.5, 
                                   'baseline_window_duration_seconds': 0.5} 

        flash_response_df = get_flash_response_df(session, response_analysis_params)
        flash_response_df = get_p_values_from_shuffled_spontaneous(session,flash_response_df)
        flash_response_df = add_image_name(session,flash_response_df)
        flash_response_df = annotate_flash_response_df_with_pref_stim(flash_response_df)

        # Test columns in flash_response_df
        for new_key in ['cell_roi_id','mean_response','baseline_response','dff_trace','dff_trace_timestamps','p_value','image_name', 'pref_stim']:
            assert new_key in flash_response_df.keys()
